# -*- coding: utf-8 -*-
"""
    emmett.html
    -----------

    Provides html generation classes.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import html
import re
import threading

from functools import reduce

from ._internal import deprecated, warnings

__all__ = ['tag', 'cat', 'safe', 'asis']


class TagStack(threading.local):
    def __init__(self):
        self.stack = []

    def __getitem__(self, key):
        return self.stack[key]

    def append(self, item):
        self.stack.append(item)

    def pop(self, idx):
        self.stack.pop(idx)

    def __bool__(self):
        return len(self.stack) > 0


class HtmlTag:
    rules = {
        'ul': ['li'],
        'ol': ['li'],
        'table': ['tr', 'thead', 'tbody'],
        'thead': ['tr'],
        'tbody': ['tr'],
        'tr': ['td', 'th'],
        'select': ['option', 'optgroup'],
        'optgroup': ['optionp']
    }
    _self_closed = {'br', 'col', 'embed', 'hr', 'img', 'input', 'link', 'meta'}

    def __init__(self, name):
        self.name = name
        self.parent = None
        self.components = []
        self.attributes = {}
        if _stack:
            _stack[-1].append(self)

    def __enter__(self):
        _stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        _stack.pop(-1)

    @staticmethod
    def wrap(component, rules):
        if rules and (
            not isinstance(component, HtmlTag) or component.name not in rules
        ):
            return HtmlTag(rules[0])(component)
        return component

    def __call__(self, *components, **attributes):
        rules = self.rules.get(self.name, [])
        self.components = [self.wrap(comp, rules) for comp in components]
        self.attributes = attributes
        for component in self.components:
            if isinstance(component, HtmlTag):
                component.parent = self
        return self

    def append(self, component):
        self.components.append(component)

    def insert(self, i, component):
        self.components.insert(i, component)

    def remove(self, component):
        self.components.remove(component)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]
        else:
            return self.attributes.get(key)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.components.insert(key, value)
        else:
            self.attributes[key] = value

    def __iter__(self):
        for item in self.components:
            yield item

    def __str__(self):
        return self.__html__()

    def __add__(self, other):
        return cat(self, other)

    def add_class(self, name):
        """ add a class to _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) | set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    def remove_class(self, name):
        """ remove a class from _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) - set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    regex_tag = re.compile(r'^([\w\-\:]+)')
    regex_id = re.compile(r'#([\w\-]+)')
    regex_class = re.compile(r'\.([\w\-]+)')
    regex_attr = re.compile(r'\[([\w\-\:]+)=(.*?)\]')

    def find(self, expr):
        union = lambda a, b: a.union(b)
        if ',' in expr:
            tags = reduce(
                union,
                [self.find(x.strip()) for x in expr.split(',')],
                set())
        elif ' ' in expr:
            tags = [self]
            for k, item in enumerate(expr.split()):
                if k > 0:
                    children = [
                        set([c for c in tag if isinstance(c, HtmlTag)])
                        for tag in tags]
                    tags = reduce(union, children)
                tags = reduce(union, [tag.find(item) for tag in tags], set())
        else:
            tags = reduce(
                union,
                [c.find(expr) for c in self if isinstance(c, HtmlTag)],
                set())
            tag = HtmlTag.regex_tag.match(expr)
            id = HtmlTag.regex_id.match(expr)
            _class = HtmlTag.regex_class.match(expr)
            attr = HtmlTag.regex_attr.match(expr)
            if (
                (tag is None or self.name == tag.group(1)) and
                (id is None or self['_id'] == id.group(1)) and
                (_class is None or _class.group(1) in
                    (self['_class'] or '').split()) and
                (attr is None or self['_' + attr.group(1)] == attr.group(2))
            ):
                tags.add(self)
        return tags

    def _build_html_attributes(self):
        return ' '.join(
            '%s="%s"' % (k[1:], k[1:] if v is True else htmlescape(v))
            for (k, v) in sorted(self.attributes.items())
            if k.startswith('_') and v is not None)

    def __html__(self):
        name = self.name
        attrs = self._build_html_attributes()
        data = self.attributes.get('data', {})
        data_attrs = ' '.join(
            'data-%s="%s"' % (k, htmlescape(v)) for k, v in data.items())
        if data_attrs:
            attrs = attrs + ' ' + data_attrs
        attrs = ' ' + attrs if attrs else ''
        if name in self._self_closed:
            return '<%s%s />' % (name, attrs)
        components = ''.join(htmlescape(v) for v in self.components)
        return '<%s%s>%s</%s>' % (name, attrs, components, name)

    def __json__(self):
        return str(self)


class MetaHtmlTag:
    def __getattr__(self, name):
        return HtmlTag(name)

    def __getitem__(self, name):
        return HtmlTag(name)


class cat(HtmlTag):
    def __init__(self, *components):
        self.components = [c for c in components]
        self.attributes = {}

    def __html__(self):
        return ''.join(htmlescape(v) for v in self.components)


class asis(HtmlTag):
    def __init__(self, text):
        self.text = text

    def __html__(self):
        return _to_str(self.text)


class safe(asis):
    @deprecated("html.safe", "html.asis")
    def __init__(self, text, sanitize=False, allowed_tags=None):
        super().__init__(text)
        if sanitize:
            warnings.warn(
                "HTML sanitizer is no longer available. "
                "Please switch to html.asis or implement your own policy."
            )
        self.sanitize = False


def _to_str(obj):
    if not isinstance(obj, str):
        return str(obj)
    return obj


def htmlescape(obj):
    if hasattr(obj, '__html__'):
        return obj.__html__()
    return html.escape(_to_str(obj), True).replace("'", "&#39;")


_stack = TagStack()
tag = MetaHtmlTag()



# -*- coding: utf-8 -*-
"""
    emmett.locals
    -------------

    Provides shortcuts to `current` object.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import Optional, cast

from pendulum import DateTime

from ._internal import ContextVarProxy as _VProxy, ObjectProxy as _OProxy
from .ctx import _ctxv, current
from .datastructures import sdict
from .language.translator import Translator
from .wrappers.request import Request
from .wrappers.response import Response
from .wrappers.websocket import Websocket

request = cast(Request, _VProxy[Request](_ctxv, 'request'))
response = cast(Response, _VProxy[Response](_ctxv, 'response'))
session = cast(Optional[sdict], _VProxy[Optional[sdict]](_ctxv, 'session'))
websocket = cast(Websocket, _VProxy[Websocket](_ctxv, 'websocket'))
T = cast(Translator, _OProxy[Translator](current, 'T'))


def now() -> DateTime:
    return current.now



# -*- coding: utf-8 -*-
"""
    tests.http
    ----------

    Test Emmett http module
"""

from helpers import current_ctx
from emmett.http import HTTP, HTTPBytes, HTTPResponse, redirect


def test_http_default():
    http = HTTP(200)

    assert http.encoded_body is b''
    assert http.status_code == 200
    assert list(http.headers) == [(b'content-type', b'text/plain')]


def test_http_bytes():
    http = HTTPBytes(200)

    assert http.body == b''
    assert http.status_code == 200
    assert list(http.headers) == [(b'content-type', b'text/plain')]


def test_http():
    response = []
    buffer = []

    def start_response(status, headers):
        response[:] = [status, headers]
        return buffer.append

    http = HTTP(
        200,
        'Hello World',
        headers={'x-test': 'Hello Header'},
        cookies={'cookie_test': 'Set-Cookie: hello cookie'}
    )

    assert http.encoded_body == b'Hello World'
    assert http.status_code == 200
    assert list(http.headers) == [
        (b'x-test', b'Hello Header'), (b'set-cookie', b'hello cookie')
    ]


def test_redirect():
    with current_ctx('/') as ctx:
        try:
            redirect('/redirect', 302)
        except HTTPResponse as http_redirect:
            assert ctx.response.status == 302
            assert http_redirect.status_code == 302
            assert list(http_redirect.headers) == [(b'location', b'/redirect')]



# -*- coding: utf-8 -*-
"""
    tests.wrappers
    --------------

    Test Emmett wrappers module
"""

from helpers import current_ctx
from emmett.testing.env import ScopeBuilder
from emmett.wrappers.request import Request
from emmett.wrappers.response import Response


def test_request():
    scope, _ = ScopeBuilder(
        path='/?foo=bar',
        method='GET',
    ).get_data()
    request = Request(scope, None, None)

    assert request.query_params == {'foo': 'bar'}
    assert request.client == '127.0.0.1'


def test_response():
    response = Response()

    assert response.status == 200
    assert response.headers['content-type'] == 'text/plain'


def test_req_ctx():
    with current_ctx('/?foo=bar') as ctx:
        assert isinstance(ctx.request, Request)
        assert isinstance(ctx.response, Response)



# -*- coding: utf-8 -*-
"""
    emmett.testing.client
    ---------------------

    Provides base classes for testing suite.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Werkzeug.
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import asyncio
import copy
import types

from io import BytesIO

from ..asgi.handlers import HTTPHandler
from ..ctx import RequestContext, current
from ..http import HTTP, HTTPResponse
from ..wrappers.request import Request
from ..wrappers.response import Response
from ..utils import cachedprop
from .env import ScopeBuilder
from .helpers import TestCookieJar, Headers
from .urls import get_host, url_parse, url_unparse


class ClientContext(object):
    def __init__(self):
        self.request = Request(current.request._scope, None, None)
        self.response = Response()
        self.response.__dict__.update(current.response.__dict__)
        self.session = copy.deepcopy(current.session)
        self.T = current.T

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass


class ClientHTTPHandler(HTTPHandler):
    async def dynamic_handler(self, scope, receive, send):
        ctx = RequestContext(
            self.app,
            scope,
            receive,
            send,
            Request,
            Response
        )
        ctx_token = current._init_(ctx)
        try:
            http = await self.router.dispatch(ctx.request, ctx.response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTP(
                    http.status_code,
                    await error_handler(),
                    headers=ctx.response.headers,
                    cookies=ctx.response.cookies
                )
        except Exception:
            self.app.log.exception('Application exception:')
            http = HTTP(
                500,
                await self.error_handler(),
                headers=ctx.response.headers
            )
        finally:
            scope['emt.ctx'] = ClientContext()
            current._close_(ctx_token)
        return http


class ClientResponse(object):
    def __init__(self, ctx, raw, status, headers):
        self.context = ctx
        self.raw = raw
        self.status = status
        self.headers = headers
        self._close = lambda: None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    @cachedprop
    def data(self):
        self._ensure_sequence()
        rv = b''.join(self.iter_encoded())
        return rv.decode('utf8')

    @property
    def is_sequence(self):
        return isinstance(self.raw, (tuple, list))

    def _ensure_sequence(self, mutable=False):
        if self.is_sequence:
            # if we need a mutable object, we ensure it's a list.
            if mutable and not isinstance(self.response, list):
                self.response = list(self.response)
            return
        self.make_sequence()

    def make_sequence(self):
        if not self.is_sequence:
            close = getattr(self.raw, 'close', None)
            self.raw = list(self.iter_encoded())
            self._close = close

    def close(self):
        if hasattr(self.raw, 'close'):
            self.response.close()
        self._close()

    def iter_encoded(self, charset='utf8'):
        for item in self.raw:
            if isinstance(item, str):
                yield item.encode(charset)
            else:
                yield item


class EmmettTestClient(object):
    """This class allows to send requests to a wrapped application."""

    def __init__(
        self, application, response_wrapper=ClientResponse, use_cookies=True,
        allow_subdomain_redirects=False
    ):
        self.application = application
        self.response_wrapper = response_wrapper
        if use_cookies:
            self.cookie_jar = TestCookieJar()
        else:
            self.cookie_jar = None
        self.allow_subdomain_redirects = allow_subdomain_redirects

    def run_asgi_app(self, scope, body):
        if self.cookie_jar is not None:
            self.cookie_jar.inject_asgi(scope)
        rv = run_asgi_app(self.application, scope, body)
        if self.cookie_jar is not None:
            self.cookie_jar.extract_asgi(scope, Headers(rv['headers']))
        return rv

    def resolve_redirect(self, response, new_loc, scope, headers):
        """Resolves a single redirect and triggers the request again
        directly on this redirect client.
        """
        scheme, netloc, script_root, qs, anchor = url_parse(new_loc)
        base_url = url_unparse((scheme, netloc, '', '', '')).rstrip('/') + '/'

        cur_name = netloc.split(':', 1)[0].split('.')
        real_name = get_host(scope, headers).rsplit(':', 1)[0].split('.')

        if len(cur_name) == 1 and not cur_name[0]:
            allowed = True
        else:
            if self.allow_subdomain_redirects:
                allowed = cur_name[-len(real_name):] == real_name
            else:
                allowed = cur_name == real_name

        if not allowed:
            raise RuntimeError('%r does not support redirect to '
                               'external targets' % self.__class__)

        status_code = response['status']
        if status_code == 307:
            method = scope['method']
        else:
            method = 'GET'

        # For redirect handling we temporarily disable the response
        # wrapper.  This is not threadsafe but not a real concern
        # since the test client must not be shared anyways.
        old_response_wrapper = self.response_wrapper
        self.response_wrapper = None
        try:
            return self.open(
                path=script_root, base_url=base_url, query_string=qs,
                method=method, as_tuple=True)
        finally:
            self.response_wrapper = old_response_wrapper

    def open(self, *args, **kwargs):
        as_tuple = kwargs.pop('as_tuple', False)
        follow_redirects = kwargs.pop('follow_redirects', False)
        scope, body = None, b''
        if not kwargs and len(args) == 1:
            if isinstance(args[0], ScopeBuilder):
                scope, body = args[0].get_data()
        if scope is None:
            builder = ScopeBuilder(*args, **kwargs)
            try:
                scope, body = builder.get_data()
            finally:
                builder.close()

        response = self.run_asgi_app(scope, body)

        # handle redirects
        redirect_chain = []
        while 1:
            status_code = response['status']
            if (
                status_code not in (301, 302, 303, 305, 307) or
                not follow_redirects
            ):
                break
            headers = Headers(response['headers'])
            new_location = headers['location']
            if new_location.startswith('/'):
                new_location = (
                    scope['scheme'] + "://" +
                    scope['server'][0] + new_location)
            new_redirect_entry = (new_location, status_code)
            if new_redirect_entry in redirect_chain:
                raise Exception('loop detected')
            redirect_chain.append(new_redirect_entry)
            scope, response = self.resolve_redirect(
                response, new_location, scope, headers)

        if self.response_wrapper is not None:
            response = self.response_wrapper(
                scope['emt.ctx'], response['body'], response['status'],
                Headers(response['headers']))
        if as_tuple:
            return scope, response
        return response

    def get(self, *args, **kw):
        """Like open but method is enforced to GET."""
        kw['method'] = 'GET'
        return self.open(*args, **kw)

    def patch(self, *args, **kw):
        """Like open but method is enforced to PATCH."""
        kw['method'] = 'PATCH'
        return self.open(*args, **kw)

    def post(self, *args, **kw):
        """Like open but method is enforced to POST."""
        kw['method'] = 'POST'
        return self.open(*args, **kw)

    def head(self, *args, **kw):
        """Like open but method is enforced to HEAD."""
        kw['method'] = 'HEAD'
        return self.open(*args, **kw)

    def put(self, *args, **kw):
        """Like open but method is enforced to PUT."""
        kw['method'] = 'PUT'
        return self.open(*args, **kw)

    def delete(self, *args, **kw):
        """Like open but method is enforced to DELETE."""
        kw['method'] = 'DELETE'
        return self.open(*args, **kw)

    def options(self, *args, **kw):
        """Like open but method is enforced to OPTIONS."""
        kw['method'] = 'OPTIONS'
        return self.open(*args, **kw)

    def trace(self, *args, **kw):
        """Like open but method is enforced to TRACE."""
        kw['method'] = 'TRACE'
        return self.open(*args, **kw)

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__,
            self.application
        )


def run_asgi_app(app, scope, body=b''):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    request_complete = False
    response_started = False
    response_complete = False
    raw = {"body": BytesIO()}

    async def receive():
        nonlocal request_complete

        if request_complete:
            while not response_complete:
                await asyncio.sleep(0.0001)
            return {"type": "http.disconnect"}

        if isinstance(body, str):
            body_bytes = body.encode("utf-8")
        elif body is None:
            body_bytes = b""
        elif isinstance(body, types.GeneratorType):
            try:
                chunk = body.send(None)
                if isinstance(chunk, str):
                    chunk = chunk.encode("utf-8")
                return {
                    "type": "http.request", "body": chunk, "more_body": True}
            except StopIteration:
                request_complete = True
                return {"type": "http.request", "body": b""}
        else:
            body_bytes = body

        request_complete = True
        return {"type": "http.request", "body": body_bytes}

    async def send(message):
        nonlocal response_started, response_complete

        if message["type"] == "http.response.start":
            raw["version"] = 11
            raw["status"] = message["status"]
            raw["headers"] = [
                (key.decode(), value.decode())
                for key, value in message["headers"]
            ]
            raw["preload_content"] = False
            response_started = True
        elif message["type"] == "http.response.body":
            body = message.get("body", b"")
            more_body = message.get("more_body", False)
            if scope['method'] != "HEAD":
                raw["body"].write(body)
            if not more_body:
                raw["body"].seek(0)
                response_complete = True

    handler = ClientHTTPHandler(app)
    loop.run_until_complete(handler(scope, receive, send))

    return raw



# -*- coding: utf-8 -*-
"""
    emmett.validators
    -----------------

    Implements validators for pyDAL.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from .basic import (
    Allow,
    Any,
    Equals,
    hasLength,
    isEmpty,
    isEmptyOr,
    isntEmpty,
    Matches,
    Not,
    Validator
)
from .consist import (
    isAlphanumeric,
    isDate,
    isDatetime,
    isDecimal,
    isEmail,
    isFloat,
    isImage,
    isInt,
    isIP,
    isJSON,
    isList,
    isTime,
    isUrl
)
from .inside import inRange, inSet, inDB, notInDB
from .process import Cleanup, Crypt, Lower, Urlify, Upper


class ValidateFromDict(object):
    numkeys = {"gt": False, "gte": True, "lt": False, "lte": True}

    def __init__(self):
        self.is_validators = {
            "alphanumeric": isAlphanumeric,
            "date": isDate,
            "datetime": isDatetime,
            "decimal": isDecimal,
            "email": isEmail,
            "float": isFloat,
            "image": isImage,
            "int": isInt,
            "ip": isIP,
            "json": isJSON,
            "time": isTime,
            "url": isUrl
        }
        self.proc_validators = {
            "clean": Cleanup,
            "crypt": Crypt,
            "lower": Lower,
            "upper": Upper,
            "urlify": Urlify
        }

    def parse_num_comparisons(self, data, minv=None, maxv=None):
        inclusions = [True, False]
        for key, include in self.numkeys.items():
            val = data.get(key)
            if val is not None:
                if key[0] == "g":
                    minv = val
                    inclusions[0] = include
                elif key[0] == "l":
                    maxv = val
                    inclusions[1] = include
        return minv, maxv, inclusions

    def parse_is(self, data, message=None):
        #: map types with fields
        if isinstance(data, str):
            #: map {'is': 'int'}
            key = data
            options = {}
        elif isinstance(data, dict):
            #: map {'is': {'float': {'dot': ','}}}
            key = list(data)[0]
            options = data[key]
        else:
            raise SyntaxError("'is' validator accepts only string or dict")
        validator = self.is_validators.get(key)
        return validator(message=message, **options) if validator else None

    def parse_is_list(self, data, message=None):
        #: map types with 'list' fields
        key = ''
        options = {}
        suboptions = {}
        lopts = ['splitter']
        if isinstance(data, str):
            #: map {'is': 'list:int'}
            key = data
        elif isinstance(data, dict):
            #: map {'is': {'list:float': {'dot': ',', 'splitter': ';'}}}
            key = list(data)[0]
            suboptions = data[key]
            for opt in list(suboptions):
                if opt in lopts:
                    options[opt] = suboptions[opt]
                    del suboptions[opt]
        else:
            raise SyntaxError("'is' validator accepts only string or dict")
        try:
            keyspecs = key.split(':')
            subkey = keyspecs[1].strip()
            assert keyspecs[0].strip() == 'list'
        except Exception:
            subkey = '_missing_'
        validator = self.is_validators.get(subkey)
        return isList(
            [validator(message=message, **suboptions)],
            message=message,
            **options
        ) if validator else None

    def parse_reference(self, field):
        ref_table = None
        multiple = None
        if field.type.startswith('reference'):
            multiple = False
        elif field.type.startswith('list:reference'):
            multiple = True
        if multiple is not None:
            ref_table = field.type.split(' ')[1]
        return ref_table, multiple

    def __call__(self, field, data):
        validators = []
        message = data.pop("message", None)
        #: parse 'presence' and 'empty'
        presence = data.get('presence')
        empty = data.get('empty')
        if presence is None and empty is not None:
            presence = not empty
        #: parse 'is'
        _is = data.get('is')
        if _is is not None:
            validator = self.parse_is(_is, message) or self.parse_is_list(_is, message)
            if validator is None:
                raise SyntaxError(
                    "Unknown type %s for 'is' validator" % data
                )
            validators.append(validator)
        #: parse 'len'
        _len = data.get('len')
        if _len is not None:
            if isinstance(_len, int):
                #: allows {'len': 2}
                validators.append(
                    hasLength(_len + 1, _len, message='Enter {min} characters')
                )
            else:
                #: allows
                #  {'len': {'gt': 1, 'gte': 2, 'lt': 5, 'lte' 6}}
                #  {'len': {'range': (2, 6)}}
                if _len.get('range') is not None:
                    minv, maxv = _len['range']
                    inc = (True, False)
                else:
                    minv, maxv, inc = self.parse_num_comparisons(_len, 0, 256)
                validators.append(hasLength(maxv, minv, inc, message=message))
        #: parse 'in'
        _dbset = None
        _in = data.get('in', [])
        if _in:
            if isinstance(_in, (list, tuple, set)):
                #: allows {'in': [1, 2]}
                validators.append(inSet(_in, message=message))
            elif isinstance(_in, dict):
                options = {}
                #: allows {'in': {'range': (1, 5)}}
                _range = _in.get('range')
                if isinstance(_range, (tuple, list)):
                    validators.append(inRange(_range[0], _range[1], message=message))
                #: allows {'in': {'set': [1, 5]}} with options
                _set = _in.get('set')
                if isinstance(_set, (list, tuple, set)):
                    opt_keys = [key for key in list(_in) if key != 'set']
                    for key in opt_keys:
                        options[key] = _in[key]
                    validators.append(inSet(_set, message=message, **options))
                #: allows {'in': {'dbset': lambda db: db.where(query)}}
                _dbset = _in.get('dbset')
                if callable(_dbset):
                    ref_table, multiple = self.parse_reference(field)
                    if ref_table:
                        opt_keys = [key for key in list(_in) if key != 'dbset']
                        for key in opt_keys:
                            options[key] = _in[key]
                        validators.append(
                            inDB(
                                field.db,
                                ref_table,
                                dbset=_dbset,
                                multiple=multiple,
                                message=message,
                                **options
                            )
                        )
                    else:
                        raise SyntaxError(
                            "'in:dbset' validator needs a reference field"
                        )
            else:
                raise SyntaxError(
                    "'in' validator accepts only a set or a dict"
                )
        #: parse 'gt', 'gte', 'lt', 'lte'
        minv, maxv, inc = self.parse_num_comparisons(data)
        if minv is not None or maxv is not None:
            validators.append(inRange(minv, maxv, inc, message=message))
        #: parse 'equals'
        if 'equals' in data:
            validators.append(Equals(data['equals'], message=message))
        #: parse 'match'
        if 'match' in data:
            if isinstance(data['match'], dict):
                validators.append(Matches(**data['match'], message=message))
            else:
                validators.append(Matches(data['match'], message=message))
        #: parse transforming validators
        for key, vclass in self.proc_validators.items():
            if key in data:
                options = {}
                if isinstance(data[key], dict):
                    options = data[key]
                elif data[key] != True:
                    if key == 'crypt' and isinstance(data[key], str):
                        options = {'algorithm': data[key]}
                    else:
                        raise SyntaxError(
                            key + " validator accepts only dict or True"
                        )
                validators.append(vclass(message=message, **options))
        #: parse 'unique'
        _unique = data.get('unique', False)
        if _unique:
            _udbset = None
            if isinstance(_unique, dict):
                whr = _unique.get('where', None)
                if callable(whr):
                    _dbset = whr
            validators.append(
                notInDB(
                    field.db,
                    field.table,
                    field.name,
                    dbset=_udbset,
                    message=message
                )
            )
            table = field.db[field._tablename]
            table._unique_fields_validation_[field.name] = 1
        #: apply 'format' option
        if 'format' in data:
            for validator in validators:
                children = [validator]
                if hasattr(validator, 'children'):
                    children += validator.children
                for child in children:
                    if hasattr(child, 'format'):
                        child.format = data['format']
                        break
        #: parse 'custom'
        if 'custom' in data:
            if isinstance(data['custom'], list):
                for element in data['custom']:
                    validators.append(element)
            else:
                validators.append(data['custom'])
        #: parse 'any'
        if 'any' in data:
            validators.append(Any(self(field, data['any']), message=message))
        #: parse 'not'
        if 'not' in data:
            validators.append(Not(self(field, data['not']), message=message))
        #: insert presence/empty validation if needed
        if presence:
            ref_table, multiple = self.parse_reference(field)
            if ref_table:
                if not _dbset:
                    validators.append(
                        inDB(field.db, ref_table, multiple=multiple, message=message)
                    )
            else:
                validators.insert(0, isntEmpty(message=message))
        if empty:
            validators.insert(0, isEmpty(message=message))
        #: parse 'allow'
        if 'allow' in data:
            if data['allow'] in ['empty', 'blank']:
                validators = [isEmptyOr(validators, message=message)]
            else:
                validators = [Allow(data['allow'], validators, message=message)]
        return validators



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.ws.websockets
    -----------------------------------

    Provides websocket websockets protocol implementation

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from uvicorn.protocols.websockets.websockets_impl import WebSocketProtocol

from . import protocols


protocols.register("websockets")(WebSocketProtocol)



# -*- coding: utf-8 -*-
"""
    emmett.asgi.helpers
    -------------------

    Provides ASGI helpers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import ssl
import sys

from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from uvicorn.config import Config as UvicornConfig
from uvicorn.lifespan.on import LifespanOn
from uvicorn.middleware.debug import DebugMiddleware
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware


class Registry:
    __slots__ = ["_data"]

    def __init__(self):
        self._data: Dict[str, Callable[..., Any]] = {}

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def keys(self) -> Iterable[str]:
        return self._data.keys()

    def register(self, key: str) -> Callable[[], Callable[..., Any]]:
        def wrap(builder: Callable[..., Any]) -> Callable[..., Any]:
            self._data[key] = builder
            return builder
        return wrap

    def get(self, key: str) -> Callable[..., Any]:
        try:
            return self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")



class BuilderRegistry(Registry):
    __slots__ = []

    def __init__(self):
        self._data: Dict[str, Tuple[Callable[..., Any], List[str]]] = {}

    def register(
        self,
        key: str,
        packages: Optional[List[str]] = None
    ) -> Callable[[], Callable[..., Any]]:
        packages = packages or []

        def wrap(builder: Callable[..., Any]) -> Callable[..., Any]:
            loaded_packages, implemented = {}, True
            try:
                for package in packages:
                    __import__(package)
                    loaded_packages[package] = sys.modules[package]
            except ImportError:
                implemented = False
            if implemented:
                self._data[key] = (builder, loaded_packages)
            return builder
        return wrap

    def get(self, key: str) -> Callable[..., Any]:
        try:
            builder, packages = self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")
        return builder(**packages)


class Config(UvicornConfig):
    def setup_event_loop(self):
        pass

    def load(self):
        assert not self.loaded

        if self.is_ssl:
            self.ssl = _create_ssl_context(
                keyfile=self.ssl_keyfile,
                certfile=self.ssl_certfile,
                cert_reqs=self.ssl_cert_reqs,
                ca_certs=self.ssl_ca_certs,
                alpn_protocols=self.http.alpn_protocols
            )
        else:
            self.ssl = None

        encoded_headers = [
            (key.lower().encode("latin1"), value.encode("latin1"))
            for key, value in self.headers
        ]
        self.encoded_headers = (
            encoded_headers if b"server" in dict(encoded_headers) else
            [(b"server", b"Emmett")] + encoded_headers
        )

        self.http_protocol_class = self.http
        self.ws_protocol_class = self.ws
        self.lifespan_class = LifespanOn

        self.loaded_app = self.app
        self.interface = "asgi3"

        if self.debug:
            self.loaded_app = DebugMiddleware(self.loaded_app)
        if self.proxy_headers:
            self.loaded_app = ProxyHeadersMiddleware(
                self.loaded_app, trusted_hosts=self.forwarded_allow_ips
            )

        self.loaded = True

    @property
    def is_ssl(self) -> bool:
        return self.ssl_certfile is not None and self.ssl_keyfile is not None


def _create_ssl_context(
    certfile: str,
    keyfile: str,
    cert_reqs: int,
    ca_certs: Optional[str],
    alpn_protocols: List[str]
) -> ssl.SSLContext:
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ctx.set_ciphers("ECDHE+AESGCM")
    ctx.options |= (
        ssl.OP_NO_SSLv2 |
        ssl.OP_NO_SSLv3 |
        ssl.OP_NO_TLSv1 |
        ssl.OP_NO_TLSv1_1 |
        ssl.OP_NO_COMPRESSION
    )
    ctx.set_alpn_protocols(alpn_protocols)
    ctx.load_cert_chain(certfile, keyfile)
    ctx.verify_mode = cert_reqs
    if ca_certs:
        ctx.load_verify_locations(ca_certs)
    return ctx



# -*- coding: utf-8 -*-
"""
    emmett.typing
    -------------

    Provides typing helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import Awaitable, Callable, TypeVar

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")

ErrorHandlerType = TypeVar("ErrorHandlerType", bound=Callable[[], Awaitable[str]])



# -*- coding: utf-8 -*-
"""
    emmett.orm.errors
    -----------------

    Provides some error wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""


class MaxConnectionsExceeded(RuntimeError):
    def __init__(self):
        super().__init__('Exceeded maximum connections')



# -*- coding: utf-8 -*-
"""
    tests.helpers
    -------------

    Tests helpers
"""

from contextlib import contextmanager

from emmett.asgi.handlers import RequestContext, WSContext
from emmett.ctx import current
from emmett.serializers import Serializers
from emmett.testing.env import ScopeBuilder
from emmett.wrappers.request import Request
from emmett.wrappers.response import Response
from emmett.wrappers.websocket import Websocket

json_dump = Serializers.get_for('json')


class FakeRequestContext(RequestContext):
    def __init__(self, app, scope):
        self.app = app
        self.request = Request(scope, None, None)
        self.response = Response()
        self.session = None


class FakeWSContext(WSContext):
    def __init__(self, app, scope):
        self.app = app
        self.websocket = Websocket(
            scope,
            self.receive,
            self.send
        )
        self._receive_storage = []
        self._send_storage = []

    async def receive(self):
        return json_dump({'foo': 'bar'})

    async def send(self, data):
        self._send_storage.append(data)


@contextmanager
def current_ctx(path, app=None):
    builder = ScopeBuilder(path)
    token = current._init_(FakeRequestContext(app, builder.get_data()[0]))
    yield current
    current._close_(token)


@contextmanager
def ws_ctx(path, app=None):
    builder = ScopeBuilder(path)
    scope_data = builder.get_data()[0]
    scope_data.update(type='websocket', scheme='wss')
    token = current._init_(FakeWSContext(app, scope_data))
    yield current
    current._close_(token)



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.engine
    ----------------------------

    Provides migration engine for pyDAL.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ...datastructures import sdict


class MetaEngine(object):
    def __init__(self, db):
        self.db = db

    def create_table(self, name, columns, **kwargs):
        self.db.create_table(name, columns)

    def drop_table(self, name):
        self.db.drop_table(name)

    def add_column(self, table_name, column):
        self.db.add_column(table_name, column)

    def drop_column(self, table_name, column_name):
        self.db.drop_column(table_name, column_name)

    def alter_column(self, table_name, column_name, changes):
        pchanges = self._parse_column_changes(changes)
        updates = {k: v[1] for k, v in pchanges.items()}
        self.db.change_column(table_name, column_name, updates)

    def create_index(self, name, table_name, fields, expr, unique, **kw):
        self.db.create_index(table_name, name, fields, expr, unique, **kw)

    def drop_index(self, name, table_name):
        self.db.drop_index(table_name, name)

    @staticmethod
    def _parse_column_changes(changes):
        rv = {}
        for change in changes:
            if change[0] == "modify_type":
                rv['type'] = [
                    change[4], change[5], change[3]['existing_length']
                ]
            elif change[0] == "modify_length":
                rv['length'] = [
                    change[4], change[5], change[3]['existing_type']
                ]
            elif change[0] == "modify_notnull":
                rv['notnull'] = [change[4], change[5]]
            elif change[0] == "modify_default":
                rv['default'] = [
                    change[4], change[5], change[3]['existing_type']
                ]
        return rv


class Engine(MetaEngine):
    @property
    def adapter(self):
        return self.db._adapter

    @property
    def dialect(self):
        return self.db._adapter.dialect

    def _log_and_exec(self, sql):
        self.db.logger.debug("executing SQL:\n%s" % sql)
        self.adapter.execute(sql)

    def create_table(self, name, columns, **kwargs):
        params = {}
        for key in ['primary_keys', 'id_col']:
            if kwargs.get(key) is not None:
                params[key] = kwargs[key]
        sql_list = self._new_table_sql(name, columns, **params)
        for sql in sql_list:
            self._log_and_exec(sql)

    def drop_table(self, name):
        adapt_v = sdict(_rname=self.dialect.quote(name))
        sql_list = self.dialect.drop_table(adapt_v, 'cascade')
        for sql in sql_list:
            self._log_and_exec(sql)

    def add_column(self, tablename, column):
        sql = self._add_column_sql(tablename, column)
        self._log_and_exec(sql)

    def drop_column(self, tablename, colname):
        sql = self._drop_column_sql(tablename, colname)
        self._log_and_exec(sql)

    def alter_column(self, table_name, column_name, changes):
        sql = self._alter_column_sql(
            table_name, column_name, self._parse_column_changes(changes))
        if sql is not None:
            self._log_and_exec(sql)

    def create_index(self, name, table_name, fields, expr, unique, **kw):
        adapt_t = sdict(_rname=self.dialect.quote(table_name))
        components = [self.dialect.quote(field) for field in fields]
        components += expr
        sql = self.dialect.create_index(
            name, adapt_t, components, unique, **kw)
        self._log_and_exec(sql)

    def drop_index(self, name, table_name):
        adapt_t = sdict(_rname=self.dialect.quote(table_name))
        sql = self.dialect.drop_index(name, adapt_t)
        self._log_and_exec(sql)

    def _gen_reference(self, tablename, column, tfks):
        referenced = column.type[10:].strip()
        constraint_name = self.dialect.constraint_name(tablename, column.name)
        try:
            rtablename, rfieldname = referenced.split('.')
        except Exception:
            rtablename = referenced
            rfieldname = 'id'
        if not rtablename:
            rtablename = tablename
        if column.fk or column.tfk:
            csql = self.adapter.types[column.type[:9]] % \
                dict(length=column.length)
            if column.fk:
                csql = csql + self.adapter.types['reference FK'] % dict(
                    constraint_name=self.dialect.quote(constraint_name),
                    foreign_key='%s (%s)' % (
                        self.dialect.quote(rtablename),
                        self.dialect.quote(rfieldname)),
                    table_name=self.dialect.quote(tablename),
                    field_name=self.dialect.quote(column.name),
                    on_delete_action=column.ondelete)
            if column.tfk:
                # TODO
                raise NotImplementedError(
                    'Migrating tables containing multiple columns references '
                    'is currently not supported.'
                )
        else:
            csql_info = dict(
                index_name=self.dialect.quote(column.name + '__idx'),
                field_name=self.dialect.quote(column.name),
                constraint_name=self.dialect.quote(constraint_name),
                foreign_key='%s (%s)' % (
                    self.dialect.quote(rtablename),
                    self.dialect.quote(rfieldname)),
                on_delete_action=column.ondelete)
            csql_info['null'] = ' NOT NULL' if column.notnull else \
                self.dialect.allow_null
            csql_info['unique'] = ' UNIQUE' if column.unique else ''
            csql = self.adapter.types['reference'] % csql_info
        return csql

    def _gen_primary_key(self, fields, primary_keys=[]):
        if primary_keys:
            fields.append(self.dialect.primary_key(
                ', '.join([
                    self.dialect.quote(pk) for pk in primary_keys])))

    def _gen_geo(self, tablename, column):
        if not hasattr(self.adapter, 'srid'):
            raise RuntimeError('Adapter does not support geometry')
        geotype, parms = column.type[:-1].split('(')
        if geotype not in self.adapter.types:
            raise SyntaxError(
                'Field: unknown field type: %s for %s' %
                (column.type, column.name))
        if self.adaper.dbengine == 'postgres' and geotype == 'geometry':
            # TODO
            raise NotImplementedError(
                'Migration with PostgreSQL and %s columns are not supported.' %
                column.type
            )
        return self.adapter.types[geotype]

    def _new_column_sql(self, tablename, column, tfks):
        if column.type.startswith('reference'):
            csql = self._gen_reference(tablename, column, tfks)
        elif column.type.startswith('list:reference'):
            csql = self.adapter.types[column.type[:14]]
        elif column.type.startswith('decimal'):
            precision, scale = map(int, column.type[8:-1].split(','))
            csql = self.adapter.types[column.type[:7]] % \
                dict(precision=precision, scale=scale)
        elif column.type.startswith('geo'):
            csql = self._gen_geo()
        elif column.type not in self.adapter.types:
            raise SyntaxError(
                'Field: unknown field type: %s for %s' %
                (column.type, column.name))
        else:
            csql = self.adapter.types[column.type] % \
                {'length': column.length}
        if self.adapter.dbengine not in ('firebird', 'informix', 'oracle'):
            cprops = "%(notnull)s%(default)s%(unique)s%(qualifier)s"
        else:
            cprops = "%(default)s%(notnull)s%(unique)s%(qualifier)s"
        if not column.type.startswith(('id', 'reference')):
            csql += cprops % {
                'notnull': ' NOT NULL' if column.notnull
                           else self.dialect.allow_null,
                'default': ' DEFAULT %s' %
                           self.adapter.represent(column.default, column.type)
                           if column.default is not None else '',
                'unique': ' UNIQUE' if column.unique else '',
                'qualifier': ' %s' % column.custom_qualifier
                             if column.custom_qualifier else ''
            }
        #     if column.notnull:
        #         csql += ' NOT NULL'
        #     else:
        #         csql += self.adapter.ALLOW_NULL()
        #     if column.unique:
        #         csql += ' UNIQUE'
        #     if column.custom_qualifier:
        #         csql += ' %s' % column.custom_qualifier
        # if column.notnull and column.default is not None:
        #     not_null = self.adapter.NOT_NULL(column.default, column.type)
        #     csql = csql.replace('NOT NULL', not_null)
        return csql

    def _new_table_sql(self, tablename, columns, primary_keys=[], id_col='id'):
        # TODO:
        # - postgres geometry
        # - SQLCustomType
        fields = []
        tfks = {}
        for sortable, column in enumerate(columns, start=1):
            csql = self._new_column_sql(tablename, column, tfks)
            fields.append('%s %s' % (self.dialect.quote(column.name), csql))
        # backend-specific extensions to fields
        if self.adapter.dbengine == 'mysql':
            if not primary_keys:
                primary_keys.append(id_col)

        self._gen_primary_key(fields, primary_keys)
        fields = ',\n    '.join(fields)
        return self.dialect.create_table(tablename, fields)

    def _add_column_sql(self, tablename, column):
        csql = self._new_column_sql(tablename, column, {})
        return 'ALTER TABLE %(tname)s ADD %(cname)s %(sql)s;' % {
            'tname': self.dialect.quote(tablename),
            'cname': self.dialect.quote(column.name),
            'sql': csql
        }

    def _drop_column_sql(self, table_name, column_name):
        if self.adapter.dbengine == "firebird":
            sql = 'ALTER TABLE %s DROP %s;'
        else:
            sql = 'ALTER TABLE %s DROP COLUMN %s;'
        return sql % (
            self.dialect.quote(table_name), self.dialect.quote(column_name))

    def _represent_changes(self, changes, field):
        if 'default' in changes and changes['default'][1] is not None:
            ftype = changes['default'][2] or field.type
            if 'type' in changes:
                ftype = changes['type'][1]
            changes['default'][1] = self.adapter.represent(
                changes['default'][1], ftype)
        if 'type' in changes:
            changes.pop('length', None)
            coltype = changes['type'][1]
            if coltype.startswith('reference'):
                raise NotImplementedError(
                    'Type change on reference fields is not supported.'
                )
            elif coltype.startswith('decimal'):
                precision, scale = map(int, coltype[8:-1].split(','))
                csql = self.adapter.types[coltype[:7]] % \
                    dict(precision=precision, scale=scale)
            elif coltype.startswith('geo'):
                csql = self._gen_geo()
            else:
                csql = self.adapter.types[coltype] % {
                    'length': changes['type'][2] or field.length
                }
            changes['type'][1] = csql
        elif 'length' in changes:
            change = changes.pop('length')
            ftype = change[2] or field.type
            changes['type'] = [
                None,
                self.adapter.types[ftype] % {'length': change[1]}
            ]

    def _alter_column_sql(self, table_name, column_name, changes):
        sql = 'ALTER TABLE %(tname)s ALTER COLUMN %(cname)s %(changes)s;'
        sql_changes_map = {
            'type': "%s" if self.adapter.dbengine in ["mysql", "mssql"] else "TYPE %s",
            'notnull': {
                True: "SET NOT NULL",
                False: "DROP NOT NULL"
            },
            'default': ["SET DEFAULT %s", "DROP DEFAULT"]
        }
        field = self.db[table_name][column_name]
        self._represent_changes(changes, field)
        sql_changes = []
        for change_type, change_val in changes.items():
            change_sql = sql_changes_map[change_type]
            if isinstance(change_sql, dict):
                sql_changes.append(change_sql[change_val[1]])
            elif isinstance(change_sql, list):
                sql_changes.append(
                    change_sql[0] % change_val[1] if change_val[1] is not None
                    else change_sql[1]
                )
            else:
                sql_changes.append(change_sql % change_val[1])
        if not sql_changes:
            return None
        return sql % {
            'tname': self.dialect.quote(table_name),
            'cname': self.dialect.quote(column_name),
            'changes': " ".join(sql_changes)
        }



# -*- coding: utf-8 -*-
"""
    emmett.orm._patches
    -------------------

    Provides pyDAL patches.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from pydal.adapters.base import BaseAdapter
from pydal.connection import ConnectionPool
from pydal.helpers.classes import ConnectionConfigurationMixin

from .adapters import (
    _initialize,
    _begin,
    _in_transaction,
    _push_transaction,
    _pop_transaction,
    _transaction_depth,
    _top_transaction
)
from .connection import (
    ConnectionManager,
    PooledConnectionManager,
    _connection_init,
    _connect_sync,
    _connect_loop,
    _close_sync,
    _close_loop,
    _connection_getter,
    _connection_setter,
    _cursors_getter,
    _connect_and_configure
)
from .engines.sqlite import SQLite


def _patch_adapter_cls():
    setattr(BaseAdapter, '_initialize_', _initialize)
    setattr(BaseAdapter, 'in_transaction', _in_transaction)
    setattr(BaseAdapter, 'push_transaction', _push_transaction)
    setattr(BaseAdapter, 'pop_transaction', _pop_transaction)
    setattr(BaseAdapter, 'transaction_depth', _transaction_depth)
    setattr(BaseAdapter, 'top_transaction', _top_transaction)
    setattr(BaseAdapter, '_connection_manager_cls', PooledConnectionManager)
    setattr(BaseAdapter, 'begin', _begin)
    setattr(SQLite, '_connection_manager_cls', ConnectionManager)


def _patch_adapter_connection():
    setattr(ConnectionPool, '__init__', _connection_init)
    setattr(ConnectionPool, 'reconnect', _connect_sync)
    setattr(ConnectionPool, 'reconnect_loop', _connect_loop)
    setattr(ConnectionPool, 'close', _close_sync)
    setattr(ConnectionPool, 'close_loop', _close_loop)
    setattr(
        ConnectionPool,
        'connection',
        property(_connection_getter, _connection_setter)
    )
    setattr(ConnectionPool, 'cursors', property(_cursors_getter))
    setattr(
        ConnectionConfigurationMixin,
        '_reconnect_and_configure',
        _connect_and_configure
)


_patch_adapter_cls()
_patch_adapter_connection()



# -*- coding: utf-8 -*-
"""
    emmett.routing.routes
    ---------------------

    Provides routes objects.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import re

from functools import wraps

import pendulum

from ..http import HTTP
from .dispatchers import (
    CacheCloseDispatcher,
    CacheDispatcher,
    CacheFlowDispatcher,
    CacheOpenDispatcher,
    Dispatcher,
    RequestCloseDispatcher,
    RequestDispatcher,
    RequestFlowDispatcher,
    RequestOpenDispatcher,
    WSCloseDispatcher,
    WSFlowDispatcher,
    WSOpenDispatcher
)

REGEX_INT = re.compile(r'<int\:(\w+)>')
REGEX_STR = re.compile(r'<str\:(\w+)>')
REGEX_ANY = re.compile(r'<any\:(\w+)>')
REGEX_ALPHA = re.compile(r'<alpha\:(\w+)>')
REGEX_DATE = re.compile(r'<date\:(\w+)>')
REGEX_FLOAT = re.compile(r'<float\:(\w+)>')


class Route:
    __slots__ = [
        'f',
        'hostname',
        'is_static',
        'match',
        'name',
        'parse_reqargs',
        'path',
        'pipeline_flow_close',
        'pipeline_flow_open',
        'regex',
        'schemes'
    ]
    _re_condl = re.compile(r'\(.*\)\?')
    _re_param = re.compile(r'<(\w+)\:(\w+)>')

    def __init__(self, rule, path, idx):
        self.name = rule.name if idx == 0 else f"{rule.name}.{idx}"
        self.f = rule.f
        if not path.startswith('/'):
            path = '/' + path
        if rule.prefix:
            path = (path != '/' and rule.prefix + path) or rule.prefix
        self.path = path
        self.schemes = tuple(rule.schemes)
        self.hostname = rule.hostname
        self.regex = re.compile(self.build_regex(self.path))
        self.pipeline_flow_open = rule.pipeline_flow_open
        self.pipeline_flow_close = rule.pipeline_flow_close
        self.build_matcher()
        self.build_argparser()

    @staticmethod
    def build_regex(path):
        path = REGEX_INT.sub(r'(?P<\g<1>>\\d+)', path)
        path = REGEX_STR.sub(r'(?P<\g<1>>[^/]+)', path)
        path = REGEX_ANY.sub(r'(?P<\g<1>>.*)', path)
        path = REGEX_ALPHA.sub(r'(?P<\g<1>>[^/\\W\\d_]+)', path)
        path = REGEX_DATE.sub(r'(?P<\g<1>>\\d{4}-\\d{2}-\\d{2})', path)
        path = REGEX_FLOAT.sub(r'(?P<\g<1>>\\d+\.\\d+)', path)
        return f'^({path})$'

    def match_simple(self, path):
        return path == self.path, {}

    def match_regex(self, path):
        match = self.regex.match(path)
        if match:
            return True, self.parse_reqargs(match)
        return False, {}

    def build_matcher(self):
        if (
            self._re_condl.findall(self.path) or
            self._re_param.findall(self.path)
        ):
            matcher, is_static = self.match_regex, False
        else:
            matcher, is_static = self.match_simple, True
        self.match = matcher
        self.is_static = is_static

    def build_argparser(self):
        parsers = {
            'int': self._parse_int_reqarg,
            'float': self._parse_float_reqarg,
            'date': self._parse_date_reqarg
        }
        opt_parsers = {
            'int': self._parse_int_reqarg_opt,
            'float': self._parse_float_reqarg_opt,
            'date': self._parse_date_reqarg_opt
        }
        pipeline = []
        for key in parsers.keys():
            optionals = []
            for element in re.compile(
                r'\(([^<]+)?<{}\:(\w+)>\)\?'.format(key)
            ).findall(self.path):
                optionals.append(element[1])
            elements = set(
                re.compile(r'<{}\:(\w+)>'.format(key)).findall(self.path))
            args = elements - set(optionals)
            if args:
                parser = self._wrap_reqargs_parser(parsers[key], args)
                pipeline.append(parser)
            if optionals:
                parser = self._wrap_reqargs_parser(
                    opt_parsers[key], optionals)
                pipeline.append(parser)
        if pipeline:
            self.parse_reqargs = self._wrap_reqargs_pipeline(pipeline)
        else:
            self.parse_reqargs = self._parse_reqargs

    @staticmethod
    def _parse_reqargs(match):
        return match.groupdict()

    @staticmethod
    def _parse_int_reqarg(args, route_args):
        for arg in args:
            route_args[arg] = int(route_args[arg])

    @staticmethod
    def _parse_int_reqarg_opt(args, route_args):
        for arg in args:
            if route_args[arg] is None:
                continue
            route_args[arg] = int(route_args[arg])

    @staticmethod
    def _parse_float_reqarg(args, route_args):
        for arg in args:
            route_args[arg] = float(route_args[arg])

    @staticmethod
    def _parse_float_reqarg_opt(args, route_args):
        for arg in args:
            if route_args[arg] is None:
                continue
            route_args[arg] = float(route_args[arg])

    @staticmethod
    def _parse_date_reqarg(args, route_args):
        try:
            for arg in args:
                route_args[arg] = pendulum.DateTime.strptime(
                    route_args[arg], "%Y-%m-%d")
        except Exception:
            raise HTTP(404)

    @staticmethod
    def _parse_date_reqarg_opt(args, route_args):
        try:
            for arg in args:
                if route_args[arg] is None:
                    continue
                route_args[arg] = pendulum.DateTime.strptime(
                    route_args[arg], "%Y-%m-%d")
        except Exception:
            raise HTTP(404)

    @staticmethod
    def _wrap_reqargs_parser(parser, args):
        @wraps(parser)
        def wrapped(route_args):
            return parser(args, route_args)
        return wrapped

    @staticmethod
    def _wrap_reqargs_pipeline(parsers):
        def wrapped(match):
            route_args = match.groupdict()
            for parser in parsers:
                parser(route_args)
            return route_args
        return wrapped

    def build_dispatcher(self, rule):
        raise NotImplementedError


class HTTPRoute(Route):
    __slots__ = ['methods', 'dispatchers']

    def __init__(self, rule, path, idx):
        super().__init__(rule, path, idx)
        self.methods = tuple(method.upper() for method in rule.methods)
        self.build_dispatchers(rule)

    def build_dispatchers(self, rule):
        dispatchers = {
            'base': (RequestDispatcher, CacheDispatcher),
            'open': (RequestOpenDispatcher, CacheOpenDispatcher),
            'close': (RequestCloseDispatcher, CacheCloseDispatcher),
            'flow': (RequestFlowDispatcher, CacheFlowDispatcher)
        }
        if self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers['flow']
        elif self.pipeline_flow_open and not self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers['open']
        elif not self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers['close']
        else:
            dispatcher, cdispatcher = dispatchers['base']
        self.dispatchers = {}
        for method in self.methods:
            dispatcher_cls = (
                cdispatcher if rule.cache_rule and method in ['HEAD', 'GET']
                else dispatcher
            )
            self.dispatchers[method] = dispatcher_cls(
                self,
                rule,
                rule.head_builder if method == 'HEAD' else rule.response_builder
            )


class WebsocketRoute(Route):
    __slots__ = ['pipeline_flow_receive', 'pipeline_flow_send', 'dispatcher']

    def __init__(self, rule, path, idx):
        super().__init__(rule, path, idx)
        self.pipeline_flow_receive = rule.pipeline_flow_receive
        self.pipeline_flow_send = rule.pipeline_flow_send
        self.build_dispatcher(rule)

    def build_dispatcher(self, rule):
        dispatchers = {
            'base': Dispatcher,
            'open': WSOpenDispatcher,
            'close': WSCloseDispatcher,
            'flow': WSFlowDispatcher
        }
        if self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher = dispatchers['flow']
        elif self.pipeline_flow_open and not self.pipeline_flow_close:
            dispatcher = dispatchers['open']
        elif not self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher = dispatchers['close']
        else:
            dispatcher = dispatchers['base']
        self.dispatcher = dispatcher(self)



# -*- coding: utf-8 -*-
"""
    emmett.app
    ----------

    Provides the central application object.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import os
import sys

from logging import Logger
from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Type, Union

import click

from yaml import SafeLoader as ymlLoader, load as ymlload

from ._internal import get_root_path, create_missing_app_folders, warn_of_deprecation
from .asgi.handlers import HTTPHandler, LifeSpanHandler, WSHandler
from .cache import RouteCacheRule
from .ctx import current
from .datastructures import sdict, ConfigData
from .extensions import Extension, ExtensionType, Signals
from .helpers import load_component
from .html import asis
from .language.helpers import Tstr
from .language.translator import Translator
from .pipeline import Pipe, Injector
from .routing.router import HTTPRouter, WebsocketRouter, RoutingCtx
from .routing.urls import url
from .templating.templater import Templater
from .testing import EmmettTestClient
from .typing import ErrorHandlerType
from .utils import dict_to_sdict, cachedprop, read_file


class Config(ConfigData):
    __slots__ = ()

    def __init__(self, app: App):
        self._app = app
        super().__init__(
            modules_class=AppModule,
            hostname_default=None,
            static_version=None,
            static_version_urls=False,
            url_default_namespace=None,
            request_max_content_length=None,
            request_body_timeout=None,
            response_timeout=None
        )
        self._handle_static = True
        self._templates_auto_reload = app.debug or False
        self._templates_encoding = 'utf8'
        self._templates_escape = 'common'
        self._templates_indent = False

    def __setattr__(self, key, value):
        obj = getattr(self.__class__, key, None)
        if isinstance(obj, property):
            return obj.fset(self, value)
        return super().__setattr__(key, value)

    @property
    def handle_static(self) -> bool:
        return self._handle_static

    @handle_static.setter
    def handle_static(self, value: bool):
        self._handle_static = value
        self._app._configure_asgi_handlers()

    @property
    def templates_auto_reload(self) -> bool:
        return self._templates_auto_reload

    @templates_auto_reload.setter
    def templates_auto_reload(self, value: bool):
        self._templates_auto_reload = value
        self._app.templater._set_reload(value)

    @property
    def templates_encoding(self) -> str:
        return self._templates_encoding

    @templates_encoding.setter
    def templates_encoding(self, value: str):
        self._templates_encoding = value
        self._app.templater._set_encoding(value)

    @property
    def templates_escape(self) -> str:
        return self._templates_escape

    @templates_escape.setter
    def templates_escape(self, value: str):
        self._templates_escape = value
        self._app.templater._set_escape(value)

    @property
    def templates_adjust_indent(self) -> bool:
        return self._templates_adjust_indent

    @templates_adjust_indent.setter
    def templates_adjust_indent(self, value: bool):
        self._templates_adjust_indent = value
        self._app.templater._set_indent(value)


class App:
    __slots__ = [
        '__dict__',
        '_asgi_handlers',
        '_extensions_env',
        '_extensions_listeners',
        '_language_default',
        '_language_force_on_url',
        '_languages_set',
        '_languages',
        '_logger',
        '_modules',
        '_pipeline',
        '_router_http',
        '_router_ws',
        'cli',
        'config_path',
        'config',
        'error_handlers',
        'ext',
        'import_name',
        'logger_name',
        'root_path',
        'static_path',
        'template_default_extension',
        'template_path',
        'templater',
        'translator'
    ]

    debug = None
    test_client_class = None

    def __init__(
        self,
        import_name: str,
        root_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        template_folder: str = 'templates',
        config_folder: str = 'config'
    ):
        self.import_name = import_name
        #: init debug var
        self.debug = os.environ.get('EMMETT_RUN_ENV') == "true"
        #: set paths for the application
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        self.static_path = os.path.join(self.root_path, "static")
        self.template_path = os.path.join(self.root_path, template_folder)
        self.config_path = os.path.join(self.root_path, config_folder)
        #: the click command line context for this application
        self.cli = click.Group(self.import_name)
        #: init the configuration
        self.config = Config(self)
        #: try to create needed folders
        create_missing_app_folders(self)
        #: init languages
        self._languages: List[str] = []
        self._languages_set: Set[str] = set()
        self._language_default: Optional[str] = None
        self._language_force_on_url = False
        self.translator = Translator(
            os.path.join(self.root_path, 'languages'),
            default_language=self.language_default or 'en',
            watch_changes=self.debug,
            str_class=Tstr
        )
        #: init routing
        self._pipeline: List[Pipe] = []
        self._router_http = HTTPRouter(self, url_prefix=url_prefix)
        self._router_ws = WebsocketRouter(self, url_prefix=url_prefix)
        self._asgi_handlers = {
            'http': HTTPHandler(self),
            'lifespan': LifeSpanHandler(self),
            'websocket': WSHandler(self)
        }
        self.error_handlers: Dict[int, Callable[[], Awaitable[str]]] = {}
        self.template_default_extension = '.html'
        #: init logger
        self._logger = None
        self.logger_name = self.import_name
        #: init extensions
        self.ext: sdict[str, Extension] = sdict()
        self._extensions_env: sdict[str, Any] = sdict()
        self._extensions_listeners: Dict[str, List[Callable[..., Any]]] = {
            element.value: [] for element in Signals
        }
        #: init templater
        self.templater: Templater = Templater(
            path=self.template_path,
            encoding=self.config.templates_encoding,
            escape=self.config.templates_escape,
            adjust_indent=self.config.templates_adjust_indent,
            reload=self.config.templates_auto_reload
        )
        #: finalise
        self._modules: Dict[str, AppModule] = {}
        current.app = self

    def _configure_asgi_handlers(self):
        self._asgi_handlers['http']._configure_methods()

    @cachedprop
    def name(self):
        if self.import_name == '__main__':
            fn = getattr(sys.modules['__main__'], '__file__', None)
            if fn is None:
                rv = '__main__'
            else:
                rv = os.path.splitext(os.path.basename(fn))[0]
        else:
            rv = self.import_name
        return rv

    @property
    def languages(self) -> List[str]:
        return self._languages

    @languages.setter
    def languages(self, value: List[str]):
        self._languages = value
        self._languages_set = set(self._languages)

    @property
    def language_default(self) -> Optional[str]:
        return self._language_default

    @language_default.setter
    def language_default(self, value: str):
        self._language_default = value
        self.translator._update_config(self._language_default or 'en')

    @property
    def language_force_on_url(self) -> bool:
        return self._language_force_on_url

    @language_force_on_url.setter
    def language_force_on_url(self, value: bool):
        self._language_force_on_url = value
        self._router_http._set_language_handling()
        self._router_ws._set_language_handling()
        self._configure_asgi_handlers()

    @property
    def pipeline(self) -> List[Pipe]:
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipes: List[Pipe]):
        self._pipeline = pipes
        self._router_http.pipeline = self._pipeline
        self._router_ws.pipeline = self._pipeline

    @property
    def injectors(self) -> List[Injector]:
        return self._router_http.injectors

    @injectors.setter
    def injectors(self, injectors: List[Injector]):
        self._router_http.injectors = injectors

    def route(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        template: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        injectors: Optional[List[Injector]] = None,
        schemes: Optional[Union[str, List[str]]] = None,
        hostname: Optional[str] = None,
        methods: Optional[Union[str, List[str]]] = None,
        prefix: Optional[str] = None,
        template_folder: Optional[str] = None,
        template_path: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        output: str = 'auto'
    ) -> RoutingCtx:
        if callable(paths):
            raise SyntaxError('Use @route(), not @route.')
        return self._router_http(
            paths=paths,
            name=name,
            template=template,
            pipeline=pipeline,
            injectors=injectors,
            schemes=schemes,
            hostname=hostname,
            methods=methods,
            prefix=prefix,
            template_folder=template_folder,
            template_path=template_path,
            cache=cache,
            output=output
        )

    def websocket(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        schemes: Optional[Union[str, List[str]]] = None,
        hostname: Optional[str] = None,
        prefix: Optional[str] = None
    ) -> RoutingCtx:
        if callable(paths):
            raise SyntaxError('Use @websocket(), not @websocket.')
        return self._router_ws(
            paths=paths,
            name=name,
            pipeline=pipeline,
            schemes=schemes,
            hostname=hostname,
            prefix=prefix
        )

    def on_error(self, code: int) -> Callable[[ErrorHandlerType], ErrorHandlerType]:
        def decorator(f: ErrorHandlerType) -> ErrorHandlerType:
            self.error_handlers[code] = f
            return f
        return decorator

    @property
    def command(self):
        return self.cli.command

    @property
    def command_group(self):
        return self.cli.group

    @property
    def log(self) -> Logger:
        if self._logger and self._logger.name == self.logger_name:
            return self._logger
        from .logger import _logger_lock, create_logger
        with _logger_lock:
            if self._logger and self._logger.name == self.logger_name:
                return self._logger
            self._logger = rv = create_logger(self)
            return rv

    def render_template(self, filename: str) -> str:
        ctx = {
            'current': current, 'url': url, 'asis': asis,
            'load_component': load_component
        }
        return self.templater.render(filename, ctx)

    def config_from_yaml(self, filename: str, namespace: Optional[str] = None):
        #: import configuration from yaml files
        rc = read_file(os.path.join(self.config_path, filename))
        rc = ymlload(rc, Loader=ymlLoader)
        c = self.config if namespace is None else self.config[namespace]
        for key, val in rc.items():
            c[key] = dict_to_sdict(val)

    #: Register modules
    def _register_module(self, mod: AppModule):
        self._modules[mod.name] = mod

    #: Creates the extensions' environments and configs
    def __init_extension(self, ext):
        if ext.namespace is None:
            ext.namespace = ext.__name__
        if self._extensions_env[ext.namespace] is None:
            self._extensions_env[ext.namespace] = sdict()
        return self._extensions_env[ext.namespace], self.config[ext.namespace]

    #: Register extension listeners
    def __register_extension_listeners(self, ext):
        for signal, listener in ext._listeners_:
            self._extensions_listeners[signal].append(listener)

    #: Add an extension to application
    def use_extension(self, ext_cls: Type[ExtensionType]) -> ExtensionType:
        if not issubclass(ext_cls, Extension):
            raise RuntimeError(
                f'{ext_cls.__name__} is an invalid Emmett extension'
            )
        ext_env, ext_config = self.__init_extension(ext_cls)
        ext = self.ext[ext_cls.__name__] = ext_cls(self, ext_env, ext_config)
        self.__register_extension_listeners(ext)
        ext.on_load()
        return ext

    #: Add a template extension to application
    def use_template_extension(self, ext_cls, **config):
        return self.templater.use_extension(ext_cls, **config)

    def send_signal(self, signal: Union[str, Signals], *args, **kwargs):
        if not isinstance(signal, Signals):
            warn_of_deprecation(
                "App.send_signal str argument",
                "extensions.Signals as argument",
                stack=3
            )
            try:
                signal = Signals[signal]
            except KeyError:
                raise SyntaxError(f"{signal} is not a valid signal")
        for listener in self._extensions_listeners[signal]:
            listener(*args, **kwargs)

    def make_shell_context(self, context: Dict[str, Any] = {}) -> Dict[str, Any]:
        context['app'] = self
        return context

    def test_client(self, use_cookies: bool = True, **kwargs) -> EmmettTestClient:
        tclass = self.test_client_class or EmmettTestClient
        return tclass(self, use_cookies=use_cookies, **kwargs)

    def __call__(self, scope, receive, send):
        return self._asgi_handlers[scope['type']](scope, receive, send)

    def module(
        self,
        import_name: str,
        name: str,
        template_folder: Optional[str] = None,
        template_path: Optional[str] = None,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        injectors: Optional[List[Injector]] = None,
        module_class: Optional[Type[AppModule]] = None,
        **kwargs: Any
    ) -> AppModule:
        module_class = module_class or self.config.modules_class
        return module_class.from_app(
            self,
            import_name,
            name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline or [],
            injectors=injectors or [],
            opts=kwargs
        )


class AppModule:
    @classmethod
    def from_app(
        cls,
        app: App,
        import_name: str,
        name: str,
        template_folder: Optional[str],
        template_path: Optional[str],
        static_folder: Optional[str],
        static_path: Optional[str],
        url_prefix: Optional[str],
        hostname: Optional[str],
        cache: Optional[RouteCacheRule],
        root_path: Optional[str],
        pipeline: List[Pipe],
        injectors: List[Injector],
        opts: Dict[str, Any] = {}
    ):
        return cls(
            app,
            name,
            import_name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline,
            injectors=injectors,
            **opts
        )

    @classmethod
    def from_module(
        cls,
        appmod: AppModule,
        import_name: str,
        name: str,
        template_folder: Optional[str],
        template_path: Optional[str],
        static_folder: Optional[str],
        static_path: Optional[str],
        url_prefix: Optional[str],
        hostname: Optional[str],
        cache: Optional[RouteCacheRule],
        root_path: Optional[str],
        opts: Dict[str, Any] = {}
    ):
        if '.' in name:
            raise RuntimeError(
                "Nested app modules' names should not contains dots"
            )
        name = appmod.name + '.' + name
        if url_prefix and not url_prefix.startswith('/'):
            url_prefix = '/' + url_prefix
        module_url_prefix = (appmod.url_prefix + (url_prefix or '')) \
            if appmod.url_prefix else url_prefix
        hostname = hostname or appmod.hostname
        cache = cache or appmod.cache
        return cls(
            appmod.app,
            name,
            import_name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=module_url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=appmod.pipeline,
            injectors=appmod.injectors,
            **opts
        )

    def module(
        self,
        import_name: str,
        name: str,
        template_folder: Optional[str] = None,
        template_path: Optional[str] = None,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        module_class: Optional[Type[AppModule]] = None,
        **kwargs: Any
    ) -> AppModule:
        module_class = module_class or self.__class__
        return module_class.from_module(
            self,
            import_name,
            name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            opts=kwargs
        )

    def __init__(
        self,
        app: App,
        name: str,
        import_name: str,
        template_folder: Optional[str] = None,
        template_path: Optional[str] = None,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        injectors: Optional[List[Injector]] = None,
        **kwargs: Any
    ):
        self.app = app
        self.name = name
        self.import_name = import_name
        if root_path is None:
            root_path = get_root_path(self.import_name)
        self.root_path = root_path
        #: - `template_folder` is referred to application `template_path`
        #  - `template_path` is referred to module root_directory unless absolute
        self.template_folder = template_folder
        if template_path and not template_path.startswith("/"):
            template_path = os.path.join(self.root_path, template_path)
        self.template_path = template_path
        #: - `static_folder` is referred to application `static_path`
        #  - `static_path` is referred to module root_directory unless absolute
        if static_path and not static_path.startswith("/"):
            static_path = os.path.join(self.root_path, static_path)
        self._static_path = (
            os.path.join(self.app.static_path, static_folder) if static_folder else
            (static_path or self.app.static_path)
        )
        self.url_prefix = url_prefix
        self.hostname = hostname
        self.cache = cache
        self._super_pipeline = pipeline or []
        self._super_injectors = injectors or []
        self.pipeline = []
        self.injectors = []
        self.app._register_module(self)

    @property
    def pipeline(self) -> List[Pipe]:
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline: List[Pipe]):
        self._pipeline = self._super_pipeline + pipeline

    @property
    def injectors(self) -> List[Injector]:
        return self._injectors

    @injectors.setter
    def injectors(self, injectors: List[Injector]):
        self._injectors = self._super_injectors + injectors

    def route(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        template: Optional[str] = None,
        **kwargs
    ) -> RoutingCtx:
        if name is not None and "." in name:
            raise RuntimeError(
                "App modules' route names should not contains dots"
            )
        name = self.name + "." + (name or "")
        pipeline = kwargs.get('pipeline', [])
        injectors = kwargs.get('injectors', [])
        if self.pipeline:
            pipeline = self.pipeline + pipeline
        kwargs['pipeline'] = pipeline
        if self.injectors:
            injectors = self.injectors + injectors
        kwargs['injectors'] = injectors
        kwargs['cache'] = kwargs.get('cache', self.cache)
        return self.app.route(
            paths=paths,
            name=name,
            template=template,
            prefix=self.url_prefix,
            template_folder=self.template_folder,
            template_path=self.template_path,
            hostname=self.hostname,
            **kwargs
        )

    def websocket(
        self,
        paths: Optional[Union[str, List[str]]] = None,
        name: Optional[str] = None,
        **kwargs
    ) -> RoutingCtx:
        if name is not None and "." in name:
            raise RuntimeError(
                "App modules' websocket names should not contains dots"
            )
        name = self.name + "." + (name or "")
        pipeline = kwargs.get('pipeline', [])
        if self.pipeline:
            pipeline = self.pipeline + pipeline
        kwargs['pipeline'] = pipeline
        return self.app.websocket(
            paths=paths,
            name=name,
            prefix=self.url_prefix,
            hostname=self.hostname,
            **kwargs
        )



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.websocket
    -------------------------

    Provides http websocket wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import Any, Dict, List, Optional, Tuple

from ..asgi.typing import Scope, Receive, Send
from . import ScopeWrapper


class Websocket(ScopeWrapper):
    __slots__ = ('_accepted', 'receive', 'send', '_flow_receive', '_flow_send')

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        super().__init__(scope, receive, send)
        self._accepted = False
        self._flow_receive = None
        self._flow_send = None
        self.receive = self._accept_and_receive
        self.send = self._accept_and_send

    def _bind_flow(self, flow_receive, flow_send):
        self._flow_receive = flow_receive
        self._flow_send = flow_send

    @property
    def _asgi_spec_version(self) -> int:
        return int(''.join(
            self._scope.get('asgi', {}).get('spec_version', '2.0').split('.')
        ))

    def _encode_headers(
        self,
        headers: Dict[str, str]
    ) -> List[Tuple[bytes, bytes]]:
        return [
            (key.encode('utf-8'), val.encode('utf-8'))
            for key, val in headers.items()
        ]

    async def accept(
        self,
        headers: Optional[Dict[str, str]] = None,
        subprotocol: Optional[str] = None
    ):
        if self._accepted:
            return
        message: Dict[str, Any] = {
            'type': 'websocket.accept',
            'subprotocol': subprotocol
        }
        if headers and self._asgi_spec_version > 20:
            message['headers'] = self._encode_headers(headers)
        await self._send(message)
        self._accepted = True
        self.receive = self._wrapped_receive
        self.send = self._wrapped_send

    async def _accept_and_receive(self) -> Any:
        await self.accept()
        return await self.receive()

    async def _accept_and_send(self, data: Any):
        await self.accept()
        await self.send(data)

    async def _wrapped_receive(self) -> Any:
        data = await self._receive()
        for method in self._flow_receive:  # type: ignore
            data = method(data)
        return data

    async def _wrapped_send(self, data: Any):
        for method in self._flow_send:  # type: ignore
            data = method(data)
        if isinstance(data, str):
            await self._send({'type': 'websocket.send', 'text': data})
        else:
            await self._send({'type': 'websocket.send', 'bytes': data})



# -*- coding: utf-8 -*-
"""
    emmett.tools.service
    --------------------

    Provides the services handler.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..ctx import current
from ..parsers import Parsers
from ..pipeline import Pipe
from ..serializers import Serializers, _json_type


class JSONServicePipe(Pipe):
    __slots__ = ['decoder', 'encoder']
    output = _json_type

    def __init__(self):
        self.decoder = Parsers.get_for('json')
        self.encoder = Serializers.get_for('json')

    async def pipe_request(self, next_pipe, **kwargs):
        current.response.headers._data['content-type'] = 'application/json'
        return self.encoder(await next_pipe(**kwargs))

    def on_receive(self, data):
        return self.decoder(data)

    def on_send(self, data):
        return self.encoder(data)


class XMLServicePipe(Pipe):
    __slots__ = ['encoder']
    output = 'str'

    def __init__(self):
        self.encoder = Serializers.get_for('xml')

    async def pipe_request(self, next_pipe, **kwargs):
        current.response.headers._data['content-type'] = 'text/xml'
        return self.encoder(await next_pipe(**kwargs))

    def on_send(self, data):
        return self.encoder(data)


def ServicePipe(procedure: str) -> Pipe:
    pipe_cls = {
        'json': JSONServicePipe,
        'xml': XMLServicePipe
    }.get(procedure)
    if not pipe_cls:
        raise RuntimeError(
            'Emmett cannot handle the service you requested: %s' %
            procedure
        )
    return pipe_cls()



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.utilities
    -------------------------------

    Provides some migration utilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ..base import Database
from .engine import Engine
from .generation import Generator
from .operations import MigrationOp, UpgradeOps


class RuntimeGenerator(Generator):
    def _load_head_to_meta(self):
        pass


class RuntimeMigration(MigrationOp):
    def __init__(self, engine: Engine, ops: UpgradeOps):
        super().__init__('runtime', ops, ops.reverse(), 'runtime')
        self.engine = engine
        for op in self.upgrade_ops.ops:
            op.engine = self.engine
        for op in self.downgrade_ops.ops:
            op.engine = self.engine

    def up(self):
        for op in self.upgrade_ops.ops:
            op.run()

    def down(self):
        for op in self.downgrade_ops.ops:
            op.run()


def generate_runtime_migration(db: Database) -> RuntimeMigration:
    engine = Engine(db)
    upgrade_ops = RuntimeGenerator.generate_from(db, None, None)
    return RuntimeMigration(engine, upgrade_ops)



# -*- coding: utf-8 -*-
"""
    emmett.__main__
    ----------------

    Alias for Emmett CLI.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

if __name__ == "__main__":
    from .cli import main

    main(as_module=True)



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.httptools
    ------------------------------------

    Provides HTTP httptools protocol implementation

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from uvicorn.protocols.http.httptools_impl import (
    HttpToolsProtocol as _HttpToolsProtocol,
    STATUS_LINE
)

from . import protocols


@protocols.register("httptools")
class HTTPToolsProtocol(_HttpToolsProtocol):
    alpn_protocols = ["http/1.1"]

    def handle_upgrade(self):
        upgrade_value = None
        for name, value in self.headers:
            if name == b"upgrade":
                upgrade_value = value.lower()

        if upgrade_value != b"websocket" or self.ws_protocol_class is None:
            msg = "Unsupported upgrade request."
            self.logger.warning(msg)

            content = [STATUS_LINE[400]]
            for name, value in self.default_headers:
                content.extend([name, b": ", value, b"\r\n"])
            content.extend(
                [
                    b"content-type: text/plain; charset=utf-8\r\n",
                    b"content-length: " + str(len(msg)).encode("ascii") + b"\r\n",
                    b"connection: close\r\n",
                    b"\r\n",
                    msg.encode("ascii"),
                ]
            )
            self.transport.write(b"".join(content))
            self.transport.close()
            return

        self.connections.discard(self)
        method = self.scope["method"].encode()
        output = [method, b" ", self.url, b" HTTP/1.1\r\n"]
        for name, value in self.scope["headers"]:
            output += [name, b": ", value, b"\r\n"]
        output.append(b"\r\n")
        protocol = self.ws_protocol_class(
            config=self.config,
            server_state=self.server_state,
            on_connection_lost=self.on_connection_lost,
        )
        protocol.connection_made(self.transport)
        protocol.data_received(b"".join(output))
        self.transport.set_protocol(protocol)
